# EFileBox.tcl --
#
#	Implements the Extended File Selection Box widget.
#
# Copyright (c) 1996, Expert Interface Technologies
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#


#
# ToDo
#   (1)	If user has entered an invalid directory, give an error dialog
#

tixWidgetClass tixExFileSelectBox2 {
    -classname TixExFileSelectBox
    -superclass tixPrimitive
    -method {
	filter invoke
    }
    -flag {
	-browsecmd -command -dialog -dir -dircmd -directory 
	-disablecallback  -pattern -selection -value
    }
    -configspec {
	{-browsecmd browseCmd BrowseCmd ""}
	{-command command Command ""}
	{-dialog dialog Dialog ""}
	{-dircmd dirCmd DirCmd ""}
	{-directory directory Directory ""}
	{-disablecallback disableCallback DisableCallback 0 tixVerifyBoolean}
	{-pattern pattern Pattern *}
	{-value value Value ""}
    }
    -alias {
	{-dir -directory}
	{-selection -value}
    }

    -default {
	{*dir.label 			{Directories:}}
	{*dir.editable 			true}
	{*dir.history 			true}
	{*dir*listbox.height 		5}
	{*file.label  			Files:}
	{*file.editable 		true}
	{*file.history 			false}
	{*file*listbox.height 		5}
	{*TixComboBox.labelSide 	top}
	{*TixComboBox*Label.anchor 	w}
	{*dir.label.underline 		0}
	{*file.label.underline		0}
	{*TixComboBox.anchor 		e}
	{*TixHList.height 		7}
	{*filelist*listbox.height 	7}
    }
}

proc tixExFileSelectBox2:InitWidgetRec {w} {
    upvar #0 $w data
    global env

    tixChainMethod $w InitWidgetRec

    if {$data(-directory) == ""} {
	global env

	if {[info exists env(PWD)]} {
	    set data(-directory) $env(PWD)
	} else {
	    set data(-directory) [pwd]
	}
    }
    set data(oldDir)    ""
    set data(flag)      0
}


#----------------------------------------------------------------------
#		Construct widget
#----------------------------------------------------------------------
proc tixExFileSelectBox2:ConstructWidget {w} {
    upvar #0 $w data

    tixChainMethod $w ConstructWidget

    # listbox frame
    set lf [frame $w.lf]

    # The pane that contains the two listboxes
    #
    set pane  [tixPanedWindow $lf.pane -orientation horizontal]
    set fpane [$pane add 1 -size 220]
    set dpane [$pane add 2 -size 220]

    $dpane config -relief flat
    $fpane config -relief flat

    # The File List Pane
    #
    set data(w:file)  [tixComboBox $fpane.file\
	-command "tixExFileSelectBox2:Cmd-FileCombo $w"\
	-prunehistory true \
	-options { \
	    label.anchor w \
	}]
    set data(w:filelist) [tixScrolledListBox $fpane.filelist \
	-command "tixExFileSelectBox2:Cmd-FileList $w 1"\
	-browsecmd "tixExFileSelectBox2:Cmd-FileList $w 0"]
    pack $data(w:file)  -padx 8 -pady 4 -side top -fill x
    pack $data(w:filelist) -padx 8 -pady 4 -side top -fill both -expand yes

    # The Directory Pane
    #
    set data(w:dir)   [tixComboBox $dpane.dir \
	-command "tixExFileSelectBox2:Cmd-DirCombo $w"\
	-prunehistory true \
	-options { \
	    label.anchor w \
	}]
    set data(w:dirlist) [tixDirList  $dpane.dirlist \
	-command "tixExFileSelectBox2:Cmd-DirList $w"\
	-browsecmd "tixExFileSelectBox2:Browse-DirList $w"]
    pack $data(w:dir)   -padx 8 -pady 4 -side top -fill x
    pack $data(w:dirlist) -padx 8 -pady 4 -side top -fill both -expand yes

    pack $pane -side top -padx 4 -pady 4 -expand yes -fill both

    pack $lf -side left -expand yes -fill both

    tixDoWhenMapped $w "tixExFileSelectBox2:Map $w"

}


#----------------------------------------------------------------------
# Configuration
#----------------------------------------------------------------------

# Update both DirList and {file list and dir combo}
#
#
proc tixExFileSelectBox2:config-directory {w value} {
    upvar #0 $w data

    if {![tixIsAbsPath $value]} {
	return $data(-directory)
    }

    set data(-directory) [tixFSAbsPath $value]
    tixSetSilent $data(w:dirlist) $data(-directory) 
    tixSetSilent $data(w:dir) $data(-directory) 
    tixWidgetDoWhenIdle tixExFileSelectBox2:LoadFiles $w reload

    return $data(-directory) 
}

proc tixExFileSelectBox2:config-filetypes {w value} {
    upvar #0 $w data

    $data(w:types) subwidget listbox delete 0 end

    foreach name [array names data] {
	if [string match type,* $name] {
	    catch {unset data($name)}
	}
    }

    if {$value == ""} {
	$data(w:types) config -state disabled
    } else {
	$data(w:types) config -state normal
    
	foreach type $value {
	    $data(w:types) insert end [lindex $type 1]
	    set data(type,[lindex $type 1]) [lindex $type 0]
	}
	tixSetSilent $data(w:types) ""
    }
}

#----------------------------------------------------------------------
# MISC Methods
#----------------------------------------------------------------------

# User activates the dir combobox
#
#
proc tixExFileSelectBox2:Cmd-DirCombo {w args} {
    upvar #0 $w data

    set dir [tixEvent flag V]
    if {![tixIsAbsPath $dir]} {
	return
    }
    set dir [tixFSAbsPath $dir]

    if {![file isdirectory $dir]} {
	return
    }

    $data(w:dirlist) config -value $dir
    set data(-directory) $dir
}

# User activates the dir list
#
#
proc tixExFileSelectBox2:Cmd-DirList {w args} {
    upvar #0 $w data

    set dir $data(-directory)
    catch {
	set dir [tixEvent flag V]
    }
    set dir [tixFSAbsPath $dir]

    tixSetSilent $data(w:dir) $dir
    set data(-directory) $dir

    tixWidgetDoWhenIdle tixExFileSelectBox2:LoadFiles $w noreload
}

# User activates the dir list
#
#
proc tixExFileSelectBox2:Browse-DirList {w args} {
    upvar #0 $w data

    set dir [tixEvent flag V]
    tixExFileSelectBox2:Cmd-DirList $w $dir
}

proc tixExFileSelectBox2:IsPattern {w string} {
    foreach char [split $string ""] {
	if {$char == "*" || $char == "?" || $char == "\{"  || $char == "\[" } {
	    return 1
	}
    }
    return 0
}

proc tixExFileSelectBox2:Cmd-FileCombo {w value} {
    upvar #0 $w data

    if {[tixEvent type] == "<Return>"} {
	tixExFileSelectBox2:Ok $w
    }
}

proc tixExFileSelectBox2:Ok {w} {
    upvar #0 $w data

    set value [string trim [$data(w:file) subwidget entry get]]
    if {$value == ""} {
	set value $data(-pattern)
    }
    tixSetSilent $data(w:file) $value

    if [tixExFileSelectBox2:IsPattern $w $value] {
	set data(-pattern) $value
	tixWidgetDoWhenIdle tixExFileSelectBox2:LoadFiles $w reload
    } else {
	if [tixIsAbsPath $value] {
	    set intName [tixFileIntName $value]
	} else {
	    set intName [tixSubFolder [tixFileIntName $data(-directory)] \
	    	[tixFileIntName $value]]
	}
	set data(-value) [tixNativeName $intName]
	tixExFileSelectBox2:Invoke $w
    }
}


proc tixExFileSelectBox2:Invoke {w} {
    upvar #0 $w data

    # Save some old history
    #
    $data(w:dir)  addhistory [$data(w:dir) cget -value]
    $data(w:file) addhistory $data(-pattern)
    $data(w:file) addhistory $data(-value)
    if {$data(-dialog) != ""} {
	eval $data(-dialog) popdown
    }
    if {$data(-command) != "" && !$data(-disablecallback)} {
	set bind(specs) "%V"
	set bind(%V) $data(-value)
	tixEvalCmdBinding $w $data(-command) bind $data(-value)
    }
}

proc tixExFileSelectBox2:Cmd-FileList {w invoke args} {
    upvar #0 $w data

    set index [lindex [$data(w:filelist) subwidget listbox curselection] 0]
    if {$index == ""} {
	set index 0
    }

    set file [$data(w:filelist) subwidget listbox get $index]
    tixSetSilent $data(w:file) $file

    set data(-value) [tixNativeName [tixSubFolder \
	[tixFileIntName $data(-directory)] [tixFileIntName $file]]]

    if {$invoke == 1} {
	tixExFileSelectBox2:Invoke $w
    } else {
	if {$data(-browsecmd) != ""} {
	    tixEvalCmdBinding $w $data(-browsecmd) "" $data(-value)
	}
    }
}

proc tixExFileSelectBox2:Cmd-TypeCombo {w args} {
    upvar #0 $w data

    set value [tixEvent flag V]

    if [info exists data(type,$value)] {
	set data(-pattern) $data(type,$value)
	tixSetSilent $data(w:file) $data(-pattern)
	tixWidgetDoWhenIdle tixExFileSelectBox2:LoadFiles $w reload
    }
}

proc tixExFileSelectBox2:LoadFiles {w flag} {
    upvar #0 $w data

    if {$flag != "reload" && $data(-directory) == $data(oldDir)} {
	return
    }

    if {![winfo ismapped [winfo toplevel $w]]} {
	tixDoWhenMapped [winfo toplevel $w] \
	    "tixExFileSelectBox2:LoadFiles $w $flag"
	return
    }

    set listbox [$data(w:filelist) subwidget listbox]
    $listbox delete 0 end

    set data(-value) ""

    tixBusy $w on [$data(w:dirlist) subwidget hlist]

    set intDir [tixFileIntName $data(-directory)]
    foreach name [tixListDir $intDir 0 1 0 \
        1 $data(-pattern)] {

	$listbox insert end [tixFileDisplayName [tixSubFolder $intDir $name]]
    }

    if {$data(oldDir) != $data(-directory)} {
	# Otherwise if the user has already selected a file and then presses
	# "show hidden", the selection won't be wiped out.
	tixSetSilent $data(w:file) $data(-pattern)
    }
    set data(oldDir) $data(-directory)

    tixWidgetDoWhenIdle tixBusy $w off [$data(w:dirlist) subwidget hlist]
}

#
# Called when thd listbox is first mapped
proc tixExFileSelectBox2:Map {w} {
    if {![winfo exists $w]} {
	return
    }

    upvar #0 $w data

    set bind(specs) "%V"
    set bind(%V) $data(-value)
    tixEvalCmdBinding $w bind \
	"tixExFileSelectBox2:Cmd-DirList $w" $data(-directory)
}

#----------------------------------------------------------------------
# Public commands
#
#----------------------------------------------------------------------
proc tixExFileSelectBox2:invoke {w} {
    tixExFileSelectBox2:Invoke $w
}

proc tixExFileSelectBox2:filter {w} {
    tixExFileSelectBox2:LoadFiles $w reload
}

